use core::{
    mem,
    sync::atomic::{AtomicU32, Ordering},
};

/// A simple lock implementation using an atomic u32.
#[repr(transparent)]
pub struct Lock {
    inner: AtomicU32,
}

impl Lock {
    const LOCKED: u32 = 0b001;
    const EMPTY: u32 = 0;

    /// Creates a new lock in the unlocked state.
    pub const fn new() -> Self {
        Self {
            inner: AtomicU32::new(0),
        }
    }

    pub fn as_ptr(&self) -> *mut u32 {
        self.inner.as_ptr()
    }

    pub unsafe fn from_ptr<'a>(ptr: *mut u32) -> &'a Self {
        // SAFETY: The caller must ensure that `ptr` is not aliased, and lasts
        // for the lifetime of the `Lock`.
        unsafe { mem::transmute(AtomicU32::from_ptr(ptr)) }
    }

    /// Acquires the lock, blocking until it is available.
    pub fn lock(&self) {
        // attempt acquiring the lock with no contention.
        if self
            .inner
            .compare_exchange_weak(
                Self::EMPTY,
                Self::LOCKED,
                Ordering::Acquire,
                Ordering::Relaxed,
            )
            .is_ok()
        {
            // We successfully acquired the lock.
            return;
        } else {
            self.lock_slow();
        }
    }

    pub fn unlock(&self) {
        // use release semantics to ensure that all previous writes are
        // available to other threads.
        self.inner.fetch_and(!Self::LOCKED, Ordering::Release);
    }

    fn lock_slow(&self) {
        // The lock is either locked, or someone is waiting for it:

        let mut spin_wait = SpinWait::new();
        let mut state = self.inner.load(Ordering::Acquire);
        loop {
            // If the lock isn't locked, we can try to acquire it.
            if state & Self::LOCKED == 0 {
                // Try to acquire the lock.
                match self.inner.compare_exchange_weak(
                    state,
                    state | Self::LOCKED,
                    Ordering::Acquire,
                    Ordering::Relaxed,
                ) {
                    Ok(_) => {
                        // We successfully acquired the lock.
                        return;
                    }
                    Err(new_state) => {
                        // The state changed, we need to check again.
                        state = new_state;
                        continue;
                    }
                }
            }

            if {
                let spun: bool;
                #[cfg(feature = "std")]
                {
                    spun = spin_wait.spin_yield();
                }
                #[cfg(not(feature = "std"))]
                {
                    spun = spin_wait.spin();
                }

                spun
            } {
                // We can spin for a little while and see if it becomes available.
                state = self.inner.load(Ordering::Relaxed);
                continue;
            }

            // If we reach here, we need to park the thread.
            atomic_wait::wait(&self.inner, Self::LOCKED);

            if self
                .inner
                .compare_exchange_weak(
                    state,
                    state | Self::LOCKED,
                    Ordering::Acquire,
                    Ordering::Relaxed,
                )
                .is_ok()
            {
                // We successfully acquired the lock after being woken up.
                return;
            }

            spin_wait.reset();
            state = self.inner.load(Ordering::Relaxed);
        }
    }

    pub fn wait(&self) {
        let state = self.inner.load(Ordering::Acquire);
        atomic_wait::wait(&self.inner, state);
    }

    pub fn wake_one(&self) {
        // Notify one thread waiting on this lock.
        atomic_wait::wake_one(&self.inner);
    }
}

// from parking_lot_core
pub struct SpinWait {
    counter: u32,
}

impl SpinWait {
    /// Creates a new `SpinWait` with an initial counter value.
    pub const fn new() -> Self {
        Self { counter: 0 }
    }

    /// Resets the counter to zero.
    pub fn reset(&mut self) {
        self.counter = 0;
    }

    pub fn spin(&mut self) -> bool {
        if self.counter >= 10 {
            // If the counter is too high, we signal the caller to potentially park.
            return false;
        }
        self.counter += 1;

        // spin for a small number of iterations based on the counter value.
        for _ in 0..(1 << self.counter) {
            core::hint::spin_loop();
        }

        true
    }

    #[cfg(feature = "std")]
    pub fn spin_yield(&mut self) -> bool {
        if self.counter >= 10 {
            // If the counter is too high, we signal the caller to potentially park.
            return false;
        }
        self.counter += 1;

        if self.counter >= 3 {
            // spin for a small number of iterations based on the counter value.
            for _ in 0..(1 << self.counter) {
                core::hint::spin_loop();
            }
        } else {
            // yield the thread and wait for the OS to reschedule us.
            std::thread::yield_now();
        }

        true
    }
}

// taken from `std`
#[derive(Debug)]
#[repr(transparent)]
pub struct Parker {
    mutex: AtomicU32,
}

impl Parker {
    const PARKED: u32 = u32::MAX;
    const EMPTY: u32 = 0;
    const NOTIFIED: u32 = 1;
    pub fn new() -> Self {
        Self {
            mutex: AtomicU32::new(Self::EMPTY),
        }
    }

    pub fn as_ptr(&self) -> *mut u32 {
        self.mutex.as_ptr()
    }

    pub unsafe fn from_ptr<'a>(ptr: *mut u32) -> &'a Self {
        // SAFETY: The caller must ensure that `ptr` is not aliased, and lasts
        // for the lifetime of the `Parker`.
        unsafe { mem::transmute(AtomicU32::from_ptr(ptr)) }
    }

    pub fn is_parked(&self) -> bool {
        self.mutex.load(Ordering::Acquire) == Self::PARKED
    }

    pub fn park(&self) {
        self.park_inner(|| ());
    }

    pub fn park_with_callback<F>(&self, before_sleep: F)
    where
        F: FnOnce(),
    {
        // This function is called when the thread is about to park.
        self.park_inner(before_sleep);
    }

    // If the caller wants to park the thread on a mutex'd condition, it is
    // possible for a deadlock to occur when the mutex is dropped just before
    // this parker atomically changes it's state to `PARKED`. For that reason,
    // we use a callback to allow the caller to perform any necessary actions
    // before parking, but after the parker is set to `PARKED`. If another
    // thread then checks the condition and checks if this thread should be
    // woken, we will be immediately notified.
    // Thusly, it is important that any caller synchronise any conditionals with
    // a mutex externally, and then unlock that mutex in `before_sleep`.
    fn park_inner<F>(&self, before_sleep: F)
    where
        F: FnOnce(),
    {
        if self.mutex.fetch_sub(1, Ordering::Acquire) == Self::NOTIFIED {
            before_sleep();
            // The thread was notified, so we can return immediately.
            return;
        }

        before_sleep();

        loop {
            atomic_wait::wait(&self.mutex, Self::PARKED);

            // We check whether we were notified or woke up spuriously with
            // acquire ordering in order to make-visible any writes made by the
            // thread that notified us.
            if self.mutex.swap(Self::EMPTY, Ordering::Acquire) == Self::NOTIFIED {
                // The thread was notified, so we can return immediately.
                return;
            } else {
                // spurious wakeup, so we need to re-park.
                continue;
            }
        }
    }

    pub fn unpark(&self) {
        // write with Release ordering to ensure that any writes made by this
        // thread are made-available to the unparked thread.
        if self.mutex.swap(Self::NOTIFIED, Ordering::Release) == Self::PARKED {
            // The thread was parked, so we need to notify it.
            atomic_wait::wake_one(&self.mutex);
        } else {
            // The thread was not parked, so we don't need to do anything.
        }
    }
}

#[cfg(feature = "alloc")]
pub mod channel {
    use alloc::sync::Arc;
    use core::{
        cell::{Cell, UnsafeCell},
        marker::PhantomData,
        mem::MaybeUninit,
        sync::atomic::{AtomicU32, Ordering},
    };

    #[repr(C)]
    #[derive(Debug)]
    struct Channel<T = ()> {
        state: AtomicU32,
        val: UnsafeCell<MaybeUninit<T>>,
    }

    unsafe impl<T: Send> Send for Channel<T> {}
    unsafe impl<T: Send> Sync for Channel<T> {}

    impl<T> Channel<T> {
        const OCCUPIED_BIT: u32 = 0b01;
        const WAITING_BIT: u32 = 0b10;

        fn new() -> Self {
            Self {
                state: AtomicU32::new(0),
                val: UnsafeCell::new(MaybeUninit::uninit()),
            }
        }
    }

    pub fn channel<T>() -> (Sender<T>, Receiver<T>) {
        let channel = Arc::new(Channel::<T>::new());
        let receiver = Receiver(channel.clone(), PhantomData);
        let sender = Sender(channel);
        (sender, receiver)
    }

    #[derive(Debug)]
    #[repr(transparent)]
    // `PhantomData<Cell<()>>` is used to ensure that `Receiver` is `!Sync` but `Send`.
    pub struct Receiver<T = ()>(Arc<Channel<T>>, PhantomData<Cell<()>>);

    #[derive(Debug)]
    #[repr(transparent)]
    pub struct Sender<T = ()>(Arc<Channel<T>>);

    impl<T> Receiver<T> {
        pub fn is_empty(&self) -> bool {
            self.0.state.load(Ordering::Acquire) & Channel::<T>::OCCUPIED_BIT == 0
        }

        pub fn as_sender(self) -> Sender<T> {
            Sender(self.0.clone())
        }

        fn wait(&mut self) {
            loop {
                let state = self
                    .0
                    .state
                    .fetch_or(Channel::<T>::WAITING_BIT, Ordering::Acquire);
                if state & Channel::<T>::OCCUPIED_BIT == 0 {
                    // The channel is empty, so we need to wait for a value to be sent.
                    // We will block until the sender wakes us up.
                    atomic_wait::wait(&self.0.state, Channel::<T>::WAITING_BIT);
                } else {
                    // The channel is occupied, so we can return.
                    self.0
                        .state
                        .fetch_and(!Channel::<T>::WAITING_BIT, Ordering::Release);
                    break;
                }
            }
        }

        /// Takes the value from the channel, if it is present.
        fn take(&mut self) -> Option<T> {
            // unset the OCCUPIED_BIT to indicate that we are taking the value, if any is present.
            if self
                .0
                .state
                .fetch_and(!Channel::<T>::OCCUPIED_BIT, Ordering::Acquire)
                & Channel::<T>::OCCUPIED_BIT
                == 0
            {
                // The channel was empty, so we return None.
                None
            } else {
                unsafe { Some(self.0.val.get().read().assume_init_read()) }
            }
        }

        pub fn recv(mut self) -> T {
            loop {
                if let Some(t) = self.take() {
                    return t;
                }

                self.wait();
            }
        }
    }

    impl<T> Sender<T> {
        pub fn send(self, value: T) {
            unsafe {
                self.0.val.get().write(MaybeUninit::new(value));
            }

            // Set the OCCUPIED_BIT to indicate that a value is present.
            let state = self
                .0
                .state
                .fetch_or(Channel::<T>::OCCUPIED_BIT, Ordering::Release);
            assert!(
                state & Channel::<T>::OCCUPIED_BIT == 0,
                "Channel is already occupied"
            );

            // If there are any receivers waiting, we need to wake them up.
            if state & Channel::<T>::WAITING_BIT != 0 {
                // There are receivers waiting, so we need to wake them up.
                atomic_wait::wake_all(&self.0.state);
            }
        }
    }
}
